package com.muyu.task.server.thread;

import java.util.concurrent.*;

public class PrioritizedThreadPool {
    private static final ExecutorService executor;
    private static final Semaphore highSemaphore;
    private static final Semaphore lowSemaphore;
    private static final Semaphore centerSemaphore;
    private static final Semaphore emergencySemaphore;

    private static final int totalThreads = 30;
    private static final int highThreads = 16;
    private static final int centerThreads = 9;
    private static final int lowThreads = 5;

    private static final int emergencyTaskThreads = 14;
    private static final int emergencyHighThreads = 9;
    private static final int emergencyCenterThreads = 5;
    private static final int emergencyLowThreads = 2;


    private static volatile boolean inEmergencyMode = false;

    static {
        if (highThreads + centerThreads + lowThreads > totalThreads) {
            throw new IllegalArgumentException("总和大于线程数");
        }
        if (emergencyCenterThreads + emergencyHighThreads + emergencyLowThreads > totalThreads) {
            throw new IllegalArgumentException("总和大于线程数");
        }
        executor =new ThreadPoolExecutor(totalThreads, totalThreads,60L,TimeUnit.SECONDS,new LinkedBlockingQueue<Runnable>());
        highSemaphore = new Semaphore(highThreads);
        centerSemaphore = new Semaphore(centerThreads);
        lowSemaphore = new Semaphore(lowThreads);
        emergencySemaphore = new Semaphore(0);

    }
    public static void submit(Runnable task) {
        adjustForEmergency();
        try {
            emergencySemaphore.acquire();
            executor.submit(()->{
                try {
                    task.run();
                }finally {
                    emergencySemaphore.release();
                    restoreDefaults();
                }
            });
        }catch (InterruptedException e){
            Thread.currentThread().interrupt();
        }
    }
    private static void subHighTask(Runnable task){
        try {
            highSemaphore.acquire();
            executor.submit(()->{
                try {
                    task.run();
                }finally {
                    highSemaphore.release();
                }
            });
        }catch (InterruptedException e){
            Thread.currentThread().interrupt();
        }
    }
    private static void subcenterTask(Runnable task){
        try {
            centerSemaphore.acquire();
            executor.submit(()->{
                try {
                    task.run();
                }finally {
                    centerSemaphore.release();
                }
            });
        }catch (InterruptedException e){
            Thread.currentThread().interrupt();
        }
    }
    private static void subLowTask(Runnable task){
        try {
            lowSemaphore.acquire();
            executor.submit(()->{
                try {
                    task.run();
                }finally {
                    lowSemaphore.release();
                }
            });
        }catch (InterruptedException e){
            Thread.currentThread().interrupt();
        }
    }

    private static synchronized void adjustForEmergency(){
        if (!inEmergencyMode) {
            inEmergencyMode = true;

            highSemaphore.release(highThreads-emergencyHighThreads);
            centerSemaphore.release(centerThreads-emergencyCenterThreads);
            lowSemaphore.release(lowThreads-emergencyLowThreads);
            emergencySemaphore.release(emergencyTaskThreads);
        }
    }

    private static synchronized void restoreDefaults(){
        if (inEmergencyMode){
            emergencySemaphore.drainPermits();
            try {
                highSemaphore.acquire(emergencyHighThreads);
                centerSemaphore.acquire(emergencyCenterThreads);
                lowSemaphore.acquire(emergencyLowThreads);
            }catch (InterruptedException e){
                throw new RuntimeException(e);
            }
        }
        highSemaphore.release(highThreads);
        centerSemaphore.release(centerThreads);
        lowSemaphore.release(lowThreads);

        emergencySemaphore.release(-emergencyTaskThreads);

        inEmergencyMode = false;
    }

    public static void  shutdown() throws InterruptedException {
        executor.shutdown();
        if (!executor.awaitTermination(1,TimeUnit.HOURS)){
            executor.shutdownNow();
            if (!executor.awaitTermination(1,TimeUnit.MINUTES)){
                System.err.println("Executor did not terminate");
            }
        }
    }




}
